from bayserver_core.bay_log import BayLog
from bayserver_core.sink import Sink
from bayserver_core.protocol.command_unpacker import CommandUnPacker

from bayserver_docker_ajp.ajp_type import AjpType
from bayserver_docker_ajp.command.cmd_data import CmdData
from bayserver_docker_ajp.command.cmd_end_response import CmdEndResponse
from bayserver_docker_ajp.command.cmd_forward_request import CmdForwardRequest
from bayserver_docker_ajp.command.cmd_get_body_chunk import CmdGetBodyChunk
from bayserver_docker_ajp.command.cmd_send_body_chunk import CmdSendBodyChunk
from bayserver_docker_ajp.command.cmd_send_headers import CmdSendHeaders
from bayserver_docker_ajp.command.cmd_shutdown import CmdShutdown

class AjpCommandUnPacker(CommandUnPacker):

    def __init__(self, handler):
        self.cmd_handler = handler
        self.reset()

    def reset(self):
        pass

    def packet_received(self, pkt):

        BayLog.debug("ajp:  packet received: type=%d data len=%d", pkt.type, pkt.data_len())

        if pkt.type == AjpType.DATA:
            cmd = CmdData()

        elif pkt.type == AjpType.FORWARD_REQUEST:
            cmd = CmdForwardRequest()

        elif pkt.type == AjpType.SEND_BODY_CHUNK:
            cmd = CmdSendBodyChunk(pkt.buf, pkt.header_len, pkt.data_len)

        elif pkt.type == AjpType.SEND_HEADERS:
            cmd = CmdSendHeaders()

        elif pkt.type == AjpType.END_RESPONSE:
            cmd = CmdEndResponse()

        elif pkt.type == AjpType.SHUTDOWN:
            cmd = CmdShutdown()

        elif pkt.type == AjpType.GET_BODY_CHUNK:
            cmd = CmdGetBodyChunk()

        else:
            raise Sink()

        cmd.unpack(pkt)
        return cmd.handle(self.cmd_handler)  # visit

    def need_data(self):
        return self.cmd_handler.need_data()



from bayserver_core.bayserver import BayServer
from bayserver_core.bay_log import BayLog
from bayserver_core.bay_message import BayMessage
from bayserver_core.symbol import Symbol
from bayserver_core.http_exception import HttpException

from bayserver_core.protocol.protocol_handler_factory import ProtocolHandlerFactory
from bayserver_core.protocol.protocol_exception import ProtocolException
from bayserver_core.agent.next_socket_action import NextSocketAction
from bayserver_core.tour.tour import Tour
from bayserver_core.tour.req_content_handler import ReqContentHandler
from bayserver_core.docker.base.inbound_handler import InboundHandler
from bayserver_core.util.http_status import HttpStatus
from bayserver_core.util.string_util import StringUtil
from bayserver_core.util.http_util import HttpUtil
from bayserver_docker_ajp.ajp_packet import AjpPacket
from bayserver_docker_ajp.ajp_protocol_handler import AjpProtocolHandler
from bayserver_docker_ajp.command.cmd_get_body_chunk import CmdGetBodyChunk
from bayserver_docker_ajp.command.cmd_send_headers import CmdSendHeaders
from bayserver_docker_ajp.command.cmd_end_response import CmdEndResponse
from bayserver_docker_ajp.command.cmd_send_body_chunk import CmdSendBodyChunk



class AjpInboundHandler(AjpProtocolHandler, InboundHandler):
    class InboundProtocolHandlerFactory(ProtocolHandlerFactory):

        def create_protocol_handler(self, pkt_store):
            return AjpInboundHandler(pkt_store)

    STATE_READ_FORWARD_REQUEST = 1
    STATE_READ_DATA = 2

    DUMMY_KEY = 1

    def __init__(self, pkt_store):
        super().__init__(pkt_store, True)
        self.cur_tour_id = None
        self.req_command = None

        self.state = None
        self.keeping = None
        self.reset_state()


    ######################################################
    # implements Reusable
    ######################################################
    def reset(self):
        super().reset()
        self.reset_state()
        self.req_command = None
        self.keeping = False
        self.cur_tour_id = 0

    ######################################################
    # implements InboundHandler
    ######################################################

    def send_res_headers(self, tur):
        chunked = False
        cmd = CmdSendHeaders()
        for name in tur.res.headers.names():
            for value in tur.res.headers.values(name):
                cmd.add_header(name, value)

        cmd.status = tur.res.headers.status
        self.command_packer.post(self.ship, cmd)

        BayLog.debug("%s send header: content-length=%d", self, tur.res.headers.content_length())

    def send_res_content(self, tur, bytes, ofs, length, callback):
        cmd = CmdSendBodyChunk(bytes, ofs, length);
        self.command_packer.post(self.ship, cmd, callback)

    def send_end_tour(self, tur, keep_alive, cb):
        BayLog.debug("%s endTour: tur=%s keep=%s", self.ship, tur, keep_alive)
        cmd = CmdEndResponse()
        cmd.reuse = keep_alive

        def ensure_func():
            if not keep_alive:
                self.command_packer.end(self.ship)

        def callback_func():
            BayLog.debug("%s call back in sendEndTour: tur=%s keep=%s", self, tur, keep_alive)
            ensure_func()
            cb()

        try:
            self.command_packer.post(self.ship, cmd, callback_func())
        except IOError as e:
            BayLog.debug("%s post failed in sendEndTour: tur=%s keep=%s", self, tur, keep_alive)
            ensure_func()
            raise e

    def send_req_protocol_error(self, e):
        tur = self.ship.get_error_tour()
        tur.res.send_error(Tour.TOUR_ID_NOCHECK, HttpStatus.BAD_REQUEST, e.message, e)
        return True



    ######################################################
    # implements AjpCommandHandler
    ######################################################
    def handle_forward_request(self, cmd):
        BayLog.debug("%s handleForwardRequest method=%s uri=%s", self.ship, cmd.method, cmd.req_uri)
        if self.state != AjpInboundHandler.STATE_READ_FORWARD_REQUEST:
            raise ProtocolException("Invalid AJP command: #{cmd.type}")

        self.keeping = False
        self.req_command = cmd
        tur = self.ship.get_tour(AjpInboundHandler.DUMMY_KEY)
        if tur is None:
            BayLog.error(BayMessage.get(Symbol.INT_NO_MORE_TOURS))
            tur = self.ship.get_tour(AjpInboundHandler.DUMMY_KEY, True)
            tur.res.send_error(Tour.TOUR_ID_NOCHECK, HttpStatus.SERVICE_UNAVAILABLE, "No available tours")
            tur.res.end_content(Tour.TOUR_ID_NOCHECK)
            self.ship.agent.shutdown(False)
            return NextSocketAction.CONTINUE

        self.cur_tour_id = tur.id
        tur.req.uri = cmd.req_uri
        tur.req.protocol = cmd.protocol
        tur.req.method = cmd.method
        cmd.headers.copy_to(tur.req.headers)
        query_string = cmd.attributes.get("?query_string")

        if StringUtil.is_set(query_string):
            tur.req.uri += "?" + query_string

        BayLog.debug("%s read header method=%s protocol=%s uri=%s contlen=%d",
                     tur, tur.req.method, tur.req.protocol, tur.req.uri, tur.req.headers.content_length())

        if BayServer.harbor.trace_header:
            for name in cmd.headers.names():
                for value in cmd.headers.values(name):
                    BayLog.info("%s header: %s=%s", tur, name, value)

        req_cont_len = cmd.headers.content_length()
        if req_cont_len > 0:
            sid = self.ship.ship_id
            def callback(le, resume):
                if resume:
                    self.ship.resume(sid)

            tur.req.set_consume_listener(req_cont_len, callback)

        try:
            self.start_tour(tur)

            if req_cont_len <= 0:
                self.end_req_content(tur)
            else:
                self.change_state(AjpInboundHandler.STATE_READ_DATA)

            return NextSocketAction.CONTINUE

        except HttpException as e:

            if req_cont_len <= 0:
                tur.res.send_http_exception(Tour.TOUR_ID_NOCHECK, e)
                self.reset_state()
                return NextSocketAction.WRITE
            else:
                # Delay send
                self.change_state(AjpInboundHandler.STATE_READ_DATA)
                tur.error = e
                tur.req.set_content_handler(ReqContentHandler.dev_null)
                return NextSocketAction.CONTINUE

    def handle_data(self, cmd):
        BayLog.debug("%s handleData len=%s", self.ship, cmd.length)

        if self.state != AjpInboundHandler.STATE_READ_DATA:
            raise RuntimeError(f"Invalid AJP command: {cmd.type} state={self.state}")

        tur = self.ship.get_tour(AjpInboundHandler.DUMMY_KEY)
        success = tur.req.post_content(Tour.TOUR_ID_NOCHECK, cmd.data, cmd.start, cmd.length)

        if tur.req.bytes_posted == tur.req.bytes_limit:
            # request content completed

            if tur.error is not None:
                tur.res.send_http_exception(Tour.TOUR_ID_NOCHECK, tur.error)
                self.reset_state()
                return NextSocketAction.WRITE
            else:
                try:
                    self.end_req_content(tur)
                    return NextSocketAction.CONTINUE
                except HttpException as e:
                    tur.res.send_http_exception(Tour.TOUR_ID_NOCHECK, e)
                    self.reset_state()
                return NextSocketAction.WRITE

        else:
            bch = CmdGetBodyChunk()
            bch.req_len = tur.req.bytes_limit - tur.req.bytes_posted
            if bch.req_len > AjpPacket.MAX_DATA_LEN:
                bch.req_len = AjpPacket.MAX_DATA_LEN

            self.command_packer.post(self.ship, bch)

            if not success:
                return NextSocketAction.SUSPEND
            else:
                return NextSocketAction.CONTINUE

    def handle_send_body_chunk(self, cmd):
        raise RuntimeError(f"Invalid AJP command: {cmd.type}")

    def handle_send_headers(self, cmd):
        raise RuntimeError(f"Invalid AJP command: {cmd.type}")

    def handle_shutdown(self, cmd):
        BayLog.info("%s handle_shutdown", self.ship)
        BayServer.shutdown()
        return NextSocketAction.CLOSE

    def handle_end_response(self, cmd):
        raise RuntimeError(f"Invalid AJP command: {cmd.type}")

    def handle_get_body_chunk(self, cmd):
        raise RuntimeError("Invalid AJP command: {cmd.type}")

    def need_data(self):
        return self.state == AjpInboundHandler.STATE_READ_DATA

    #
    # private
    #

    def reset_state(self):
        self.change_state(AjpInboundHandler.STATE_READ_FORWARD_REQUEST)

    def change_state(self, new_state):
        self.state = new_state

    def end_req_content(self, tur):
        tur.req.end_content(Tour.TOUR_ID_NOCHECK)
        self.reset_state()

    def start_tour(self, tur):
        HttpUtil.parse_host_port(tur, 443 if self.req_command.is_ssl else 80)
        HttpUtil.parse_authorization(tur)

        skt = self.ship.socket

        tur.req.remote_port = None
        tur.req.remote_address = self.req_command.remote_addr

        tur.req.remote_host_func = lambda: self.req_command.remote_host

        server_addr = skt.getsockname()
        tur.req.server_address = server_addr[0]

        tur.req.server_port = self.req_command.server_port
        tur.req.server_name = self.req_command.server_name
        tur.is_secure = self.req_command.is_ssl

        tur.go()


from bayserver_core.protocol.packet import Packet
from bayserver_core.protocol.packet_part_accessor import PacketPartAccessor
from bayserver_core.util.string_util import StringUtil


#
#  AJP Protocol
#  https://tomcat.apache.org/connectors-doc/ajp/ajpv13a.html
#
#  AJP packet spec
#
#    packet:  preamble, length, body
#    preamble:
#         0x12, 0x34  (client->server)
#      | 'A', 'B'     (server->client)
#    length:
#       2 byte
#    body:
#       $length byte
#
#
#   Body format
#     client->server
#     Code     Type of Packet    Meaning
#        2     Forward Request   Begin the request-processing cycle with the following data
#        7     Shutdown          The web server asks the container to shut itself down.
#        8     Ping              The web server asks the container to take control (secure login phase).
#       10     CPing             The web server asks the container to respond quickly with a CPong.
#     none     Data              Size (2 bytes) and corresponding body data.
#
#     server->client
#     Code     Type of Packet    Meaning
#        3     Send Body Chunk   Send a chunk of the body from the servlet container to the web server (and presumably, onto the browser).
#        4     Send Headers      Send the response headers from the servlet container to the web server (and presumably, onto the browser).
#        5     End Response      Marks the end of the response (and thus the request-handling cycle).
#        6     Get Body Chunk    Get further data from the request if it hasn't all been transferred yet.
#        9     CPong Reply       The reply to a CPing request
#
#

class AjpPacket(Packet):

    class AjpAccessor(PacketPartAccessor):
        def __init__(self, pkt, start, max_len):
            super().__init__(pkt, start, max_len)

        def put_string(self, s):
            if StringUtil.is_empty(s):
                self.put_short(0xffff)
            else:
                self.put_short(len(s))
                super().put_string(s)
                self.put_byte(0)  # null terminator

        def get_string(self):
            return self.get_string_by_len(self.get_short())

        def get_string_by_len(self, length):

            if length == 0xffff:
                return ""

            buf = bytearray(length)
            self.get_bytes(buf, 0, length)
            self.get_byte()  # null terminator

            return StringUtil.from_bytes(buf)

    PREAMBLE_SIZE = 4
    MAX_DATA_LEN = 8192 - PREAMBLE_SIZE
    MIN_BUF_SIZE = 1024

    def __init__(self, typ):
        super().__init__(typ, AjpPacket.PREAMBLE_SIZE, AjpPacket.MAX_DATA_LEN)
        self.to_server = False

    def __str__(self):
        return f"AjpPacket({self.type})"

    def reset(self):
        super().reset()
        self.to_server = False

    def new_ajp_header_accessor(self):
        return AjpPacket.AjpAccessor(self, 0, AjpPacket.PREAMBLE_SIZE)

    def new_ajp_data_accessor(self):
        return AjpPacket.AjpAccessor(self, AjpPacket.PREAMBLE_SIZE, -1)




from bayserver_core.agent.transporter.plain_transporter import PlainTransporter
from bayserver_core.protocol.packet_store import PacketStore
from bayserver_core.protocol.protocol_handler_store import ProtocolHandlerStore
from bayserver_core.docker.warp.warp_docker import WarpDocker
from bayserver_core.util.io_util import IOUtil

from bayserver_docker_ajp.ajp_docker import AjpDocker
from bayserver_docker_ajp.ajp_packet_factory import AjpPacketFactory
from bayserver_docker_ajp.ajp_warp_handler import AjpWarpHandler

class AjpWarpDocker(WarpDocker, AjpDocker):

    ######################################################
    # Implements WarpDocker
    ######################################################
    def secure(self):
        return False

    ######################################################
    # Implements WarpDockerBase
    ######################################################
    def protocol(self):
        return AjpDocker.PROTO_NAME

    def new_transporter(self, agt, skt):
        return PlainTransporter(False, IOUtil.get_sock_recv_buf_size(skt))

    ######################################################
    # Class initializer
    ######################################################
    PacketStore.register_protocol(
        AjpDocker.PROTO_NAME,
        AjpPacketFactory()
    )
    ProtocolHandlerStore.register_protocol(
        AjpDocker.PROTO_NAME,
        False,
        AjpWarpHandler.WarpProtocolHandlerFactory())


from bayserver_docker_ajp.ajp_command import AjpCommand
from bayserver_docker_ajp.ajp_type import AjpType


#
#  End response body format
#
#  AJP13_END_RESPONSE :=
#    prefix_code       5
#    reuse             (boolean)
#

class CmdEndResponse(AjpCommand):

    def __init__(self):
        super().__init__(AjpType.END_RESPONSE, False)
        self.reuse = None

    def pack(self, pkt):
        acc = pkt.new_ajp_data_accessor()
        acc.put_byte(self.type)
        acc.put_byte(1 if self.reuse else 0)

        #  must be called from last line
        super().pack(pkt)

    def unpack(self, pkt):
        super().unpack(pkt)
        acc = pkt.new_ajp_data_accessor()
        acc.get_byte()  # prefix code
        self.reuse = acc.get_byte() != 0


    def handle(self, handler):
        return handler.handle_end_response(self)



from bayserver_core.docker.base.port_base import PortBase
from bayserver_core.protocol.packet_store import PacketStore
from bayserver_core.protocol.protocol_handler_store import ProtocolHandlerStore

from bayserver_docker_ajp.ajp_docker import AjpDocker
from bayserver_docker_ajp.ajp_packet_factory import AjpPacketFactory
from bayserver_docker_ajp.ajp_inbound_handler import AjpInboundHandler


class AjpPortDocker(PortBase, AjpDocker):

    ######################################################
    # Implements Port
    ######################################################
    def protocol(self):
        return AjpDocker.PROTO_NAME


    ######################################################
    # Implements PortBase
    ######################################################
    def support_anchored(self):
        return True

    def support_unanchored(self):
        return False

    ######################################################
    # Class initializer
    ######################################################

    PacketStore.register_protocol(
        AjpDocker.PROTO_NAME,
        AjpPacketFactory()
    )
    ProtocolHandlerStore.register_protocol(
        AjpDocker.PROTO_NAME,
        True,
        AjpInboundHandler.InboundProtocolHandlerFactory())



from bayserver_core.protocol.command import Command
from bayserver_core.util.char_util import CharUtil

class AjpCommand(Command):

    def __init__(self, typ, to_server):
        super().__init__(typ)
        self.to_server = to_server

    def unpack(self, pkt):
        if pkt.type != self.type:
            raise RuntimeError("Illegal State")

        self.to_server = pkt.to_server

    #
    # Super class method must be called from last line of override method
    # since header cannot be packed before data is constructed.
    #
    def pack(self, pkt):
        if pkt.type != self.type:
            raise RuntimeError("Illegal State")

        pkt.to_server = self.to_server
        self.pack_header(pkt)

    def pack_header(self, pkt):
        acc = pkt.new_ajp_header_accessor()
        if pkt.to_server:
            acc.put_byte(0x12)
            acc.put_byte(0x34)
        else:
            acc.put_byte(CharUtil.A_BYTE)
            acc.put_byte(CharUtil.B_BYTE)

        acc.put_byte((pkt.data_len() >> 8) & 0xff)
        acc.put_byte(pkt.data_len() & 0xff)



from bayserver_core.protocol.protocol_handler import ProtocolHandler
from bayserver_core.protocol.packet_packer import PacketPacker
from bayserver_core.protocol.command_packer import CommandPacker

from bayserver_docker_ajp.ajp_command_unpacker import AjpCommandUnPacker
from bayserver_docker_ajp.ajp_docker import AjpDocker
from bayserver_docker_ajp.ajp_packet_unpacker import AjpPacketUnPacker
from bayserver_docker_ajp.ajp_command_handler import AjpCommandHandler
from bayserver_docker_ajp.command.cmd_data import CmdData
from bayserver_docker_ajp.command.cmd_send_body_chunk import CmdSendBodyChunk


class AjpProtocolHandler(ProtocolHandler, AjpCommandHandler):

    def __init__(self, pkt_store, svr_mode):
        super().__init__()
        self.command_unpacker = AjpCommandUnPacker(self)
        self.packet_unpacker = AjpPacketUnPacker(pkt_store, self.command_unpacker)
        self.packet_packer = PacketPacker()
        self.command_packer = CommandPacker(self.packet_packer, pkt_store)
        self.server_mode = svr_mode

    def __str__(self):
        return f"pch[{self.ship}]"


    ######################################################
    # Implements ProtocolHandler
    ######################################################

    def protocol(self):
        return AjpDocker.PROTO_NAME

    def max_req_packet_data_size(self):
        return CmdData.MAX_DATA_LEN

    def max_res_packet_data_size(self):
        return CmdSendBodyChunk.MAX_CHUNKLEN




from bayserver_docker_ajp.ajp_command import AjpCommand
from bayserver_docker_ajp.ajp_type import AjpType

#
# Get Body Chunk format
#
# AJP13_GET_BODY_CHUNK :=
#   prefix_code       6
#   requested_length  (integer)
#

class CmdGetBodyChunk(AjpCommand):

    def __init__(self):
        super().__init__(AjpType.GET_BODY_CHUNK, False)
        self.req_len = None

    def pack(self, pkt):
        acc = pkt.new_ajp_data_accessor()
        acc.put_byte(self.type)
        acc.put_short(self.req_len)

        # must be called from last line
        super().pack(pkt)

    def handle(self, handler):
        return handler.handle_get_body_chunk(self)



from bayserver_docker_ajp.ajp_command import AjpCommand
from bayserver_docker_ajp.ajp_packet import AjpPacket
from bayserver_docker_ajp.ajp_type import AjpType

#
#  Data command format
#
#  AJP13_DATA :=
#    len, raw data
#

class CmdData(AjpCommand):
    LENGTH_OF_SIZE = 2
    MAX_DATA_LEN = AjpPacket.MAX_DATA_LEN - LENGTH_OF_SIZE

    def __init__(self, data=None, start=0, length=0):
        super().__init__(AjpType.DATA, True)
        self.data = data
        self.start = start
        self.length = length

    def unpack(self, pkt):
        super().unpack(pkt)
        acc = pkt.new_ajp_data_accessor()
        self.length = acc.get_short()
        self.data = pkt.buf
        self.start = pkt.header_len + 2

    def pack(self, pkt):
        acc = pkt.new_ajp_data_accessor()
        acc.put_short(self.length)
        acc.put_bytes(self.data, self.start, self.length)

        # BayLog.debug "pack AJP command data: #{pkt.data.bytes}"

        #  must be called from last line
        super().pack(pkt)

    def handle(self, handler):
        return handler.handle_data(self)


from bayserver_core.bayserver import BayServer
from bayserver_core.bay_log import BayLog
from bayserver_core.protocol.protocol_exception import ProtocolException
from bayserver_core.agent.next_socket_action import NextSocketAction
from bayserver_core.tour.tour import Tour
from bayserver_core.docker.warp.warp_data import WarpData
from bayserver_core.docker.warp.warp_handler import WarpHandler
from bayserver_docker_ajp.ajp_protocol_handler import AjpProtocolHandler
from bayserver_docker_ajp.command.cmd_forward_request import CmdForwardRequest
from bayserver_docker_ajp.command.cmd_data import CmdData


class AjpWarpHandler(AjpProtocolHandler, WarpHandler):
    class WarpProtocolHandlerFactory:

        def create_protocol_handler(self, pkt_store):
            return AjpWarpHandler(pkt_store)

    FIXED_WARP_ID = 1

    STATE_READ_HEADER = 1
    STATE_READ_CONTENT = 2

    def __init__(self, pkt_store):
        super().__init__(pkt_store, False)
        self.state = None
        self.cont_read_len = None
        self.reset()

    def reset(self):
        super().reset()
        self.reset_state()
        self.cont_read_len = 0


    def __str__(self):
        return str(self.ship)


    ######################################################
    # Implements WarpHandler
    ######################################################
    def next_warp_id(self):
        return 1

    def new_warp_data(self, warp_id):
        return WarpData(self.ship, warp_id)

    def post_warp_headers(self, tur):
        self.send_forward_request(tur)

    def post_warp_contents(self, tur, buf, start, length, callback):
        self.send_data(tur, buf, start, length, callback)

    def post_warp_end(self, tur):
        pass

    def verify_protocol(self, proto):
        pass

    ######################################################
    # Implements AjpCommandHandler
    ######################################################
    def handle_data(self, cmd):
        raise ProtocolException("Invalid AJP command: %d", cmd.type)

    def handle_end_response(self, cmd):
        BayLog.debug("%s handle_end_response reuse=%s", self.ship, cmd.reuse)
        tur = self.ship.get_tour(AjpWarpHandler.FIXED_WARP_ID)

        if self.state == AjpWarpHandler.STATE_READ_HEADER:
            self.end_res_header(tur)

        self.end_res_content(tur)
        if cmd.reuse:
            return NextSocketAction.CONTINUE
        else:
            return NextSocketAction.CLOSE

    def handle_forward_request(self, cmd):
        raise ProtocolException("Invalid AJP command: %s", cmd.type)

    def handle_send_body_chunk(self, cmd):
        BayLog.debug("%s handle_send_body_chunk len=%d", self.ship, cmd.length)
        tur = self.ship.get_tour(AjpWarpHandler.FIXED_WARP_ID)

        if self.state == AjpWarpHandler.STATE_READ_HEADER:

            sid = self.ship.ship_id

            def callback(length, resume):
                if resume:
                    self.ship.resume(sid)

            tur.res.set_consume_listener(callback)

            self.end_res_header(tur)

        available = tur.res.send_content(tur.tour_id, cmd.chunk, 0, cmd.length)
        self.cont_read_len += cmd.length

        if available:
            return NextSocketAction.CONTINUE
        else:
            return NextSocketAction.SUSPEND

    def handle_send_headers(self, cmd):
        BayLog.debug("%s handle_send_headers", self.ship)

        tur = self.ship.get_tour(AjpWarpHandler.FIXED_WARP_ID)

        if self.state != AjpWarpHandler.STATE_READ_HEADER:
            raise ProtocolException("Invalid AJP command: %d state=%s", cmd.type, self.state)

        wdat = WarpData.get(tur)

        if BayServer.harbor.trace_header:
            BayLog.info("%s recv res status: %d", wdat, cmd.status)

        wdat.res_headers.status = cmd.status
        for name in cmd.headers.keys():
            for value in cmd.headers[name]:
                if BayServer.harbor.trace_header:
                    BayLog.info("%s recv res header: %s=%s", wdat, name, value)

                wdat.res_headers.add(name, value)


        return NextSocketAction.CONTINUE

    def handle_shutdown(self, cmd):
        raise ProtocolException("Invalid AJP command: %d", cmd.type)

    def handle_get_body_chunk(self, cmd):
        BayLog.debug("%s handle_get_body_chunk", self)
        return NextSocketAction.CONTINUE

    def handle_eof(self):
        raise EOFError()

    def need_data(self):
        return False

    #
    # private
    #
    def end_res_header(self, tur):
        wdat = WarpData.get(tur)
        wdat.res_headers.copy_to(tur.res.headers)
        tur.res.send_headers(Tour.TOUR_ID_NOCHECK)
        self.change_state(AjpWarpHandler.STATE_READ_CONTENT)

    def end_res_content(self, tur):
        self.ship.end_warp_tour(tur)
        tur.res.end_content(Tour.TOUR_ID_NOCHECK)
        self.reset_state()

    def change_state(self, new_state):
        self.state = new_state

    def reset_state(self):
        self.change_state(AjpWarpHandler.STATE_READ_HEADER)

    def send_forward_request(self, tur):
        BayLog.debug("%s construct header", tur)

        cmd = CmdForwardRequest()
        cmd.to_server = True
        cmd.method = tur.req.method
        cmd.protocol = tur.req.protocol
        if tur.req.rewritten_uri:
            rel_uri = tur.req.rewritten_uri
        else:
            rel_uri = tur.req.uri
        town_path = tur.town.name
        if not town_path.endswith("/"):
            town_path += "/"

        rel_uri = rel_uri[len(town_path):]
        req_uri = self.ship.docker.warp_base + rel_uri

        pos = req_uri.find('?')
        if pos > 0:
            cmd.req_uri = req_uri[0:pos]
            cmd.attributes["?query_string"] = req_uri[pos + 1:]
        else:
            cmd.req_uri = req_uri

        cmd.remote_addr = tur.req.remote_address
        cmd.remote_host = tur.req.remote_host()
        cmd.server_name = tur.req.server_name
        cmd.server_port = self.ship.docker.port
        cmd.is_ssl = tur.is_secure
        cmd.headers = tur.req.headers

        if BayServer.harbor.trace_header:
            for name in cmd.headers.names():
                for value in cmd.headers.values(name):
                    BayLog.info("%s sendWarpHeader: %s=%s", WarpData.get(tur), name, value)


        self.command_packer.post(self.ship, cmd)

    def send_data(self, tur, data, ofs, length, callback):
        BayLog.debug("%s construct contents", tur)

        cmd = CmdData(data, ofs, length)
        cmd.to_server = True

        self.command_packer.post(self.ship, cmd, callback)
